@@ -5,7 +5,7 @@
 
 extern crate alloc;
 
-mod build;
+mod builder;
 mod inline;
 
 use alloc::rc::Rc;
@@ -243,6 +243,23 @@ macro_rules! flexstr {
                 }
             }
 
+            impl From<builder::FlexStrBuilder> for $name {
+                #[inline]
+                fn from(builder: builder::FlexStrBuilder) -> Self {
+                    match builder {
+                        builder::FlexStrBuilder::Small(buffer) => {
+                            let len: u8 = buffer.len() as u8;
+                            $name([<$name Inner>]::Inlined(inline::InlineFlexStr::from_array(
+                                buffer.into_inner(),
+                                len,
+                            )))
+                        }
+                        builder::FlexStrBuilder::Regular(buffer) => buffer.[<to_ $lower_name>](),
+                        builder::FlexStrBuilder::Large(s) => s.into(),
+                    }
+                }
+            }
+
             #[doc = "Converts a `String` into a `" $name "`"]
             /// ```
             #[doc = "use flexstr::" $name ";"]
@@ -364,7 +381,7 @@ macro_rules! flexstr {
 
             // TODO: Is there value in making this a public macro with varargs? Hmm...
             fn [<$lower_name _concat>](s1: &str, s2: &str) -> $name {
-                let mut builder = build::FlexStrBuilder::with_capacity(s1.len() + s2.len());
+                let mut builder = builder::FlexStrBuilder::with_capacity(s1.len() + s2.len());
                 unsafe {
                     // Safety: write_str always succeeds
                     builder.write_str(s1).unwrap_unchecked();
@@ -462,6 +479,86 @@ macro_rules! flexstr {
                 }
             }
 
+            impl [<Into $name>] for builder::FlexStrBuilder {
+                #[inline]
+                fn [<into_ $lower_name>](self) -> $name {
+                    self.into()
+                }
+            }
+
+            // *** ToCase custom trait ***
+
+            #[doc = "Trait that provides uppercase/lowercase conversion functions for `" $name "`"]
+            pub trait [<ToCase $name>] {
+                #[doc = "Converts string to uppercase and returns a `" $name "`"]
+                fn to_uppercase(&self) -> $name;
+
+                #[doc = "Converts string to lowercase and returns a `" $name "`"]
+                fn to_lowercase(&self) -> $name;
+
+                #[doc = "Converts string to ASCII uppercase and returns a `" $name "`"]
+                fn to_ascii_uppercase(&self) -> $name;
+
+                #[doc = "Converts string to ASCII lowercase and returns a `" $name "`"]
+                fn to_ascii_lowercase(&self) -> $name;
+            }
+
+            impl [<ToCase $name>] for str {
+                fn to_uppercase(&self) -> $name {
+                    // We estimate capacity based on previous string, but if not ASCII this might be wrong
+                    let mut builder = builder::FlexStrBuilder::with_capacity(self.len());
+
+                    for ch in self.chars() {
+                        let upper_chars = ch.to_uppercase();
+                        for ch in upper_chars {
+                            // Safety: Wraps `write_str` which always succeeds
+                            unsafe { builder.write_char(ch).unwrap_unchecked() }
+                        }
+                    }
+
+                    builder.into()
+                }
+
+                fn to_lowercase(&self) -> $name {
+                    // We estimate capacity based on previous string, but if not ASCII this might be wrong
+                    let mut builder = builder::FlexStrBuilder::with_capacity(self.len());
+
+                    for ch in self.chars() {
+                        let lower_chars = ch.to_lowercase();
+                        for ch in lower_chars {
+                            // Safety: Wraps `write_str` which always succeeds
+                            unsafe { builder.write_char(ch).unwrap_unchecked() }
+                        }
+                    }
+
+                    builder.into()
+                }
+
+                fn to_ascii_uppercase(&self) -> $name {
+                    let mut builder = builder::FlexStrBuilder::with_capacity(self.len());
+
+                    for mut ch in self.chars() {
+                        char::make_ascii_uppercase(&mut ch);
+                        // Safety: Wraps `write_str` which always succeeds
+                        unsafe { builder.write_char(ch).unwrap_unchecked() }
+                    }
+
+                    builder.into()
+                }
+
+                fn to_ascii_lowercase(&self) -> $name {
+                    let mut builder = builder::FlexStrBuilder::with_capacity(self.len());
+
+                    for mut ch in self.chars() {
+                        char::make_ascii_lowercase(&mut ch);
+                        // Safety: Wraps `write_str` which always succeeds
+                        unsafe { builder.write_char(ch).unwrap_unchecked() }
+                    }
+
+                    builder.into()
+                }
+            }
+
             // *** Optional serialization support ***
 
             #[cfg(feature = "serde")]
@@ -532,7 +629,7 @@ pub fn flex_fmt(args: Arguments<'_>) -> FlexStr {
     // NOTE: We have a disadvantage to `String` because we cannot call `estimated_capacity()` on args
     // As such, we cannot assume a given needed capacity - we start with a stack allocated buffer
     // and only promote to a heap buffer if a write won't fit
-    let mut builder = build::FlexStrBuilder::new();
+    let mut builder = builder::FlexStrBuilder::new();
     builder
         .write_fmt(args)
         .expect("a formatting trait implementation returned an error");
@@ -561,7 +658,7 @@ pub fn a_flex_fmt(args: Arguments<'_>) -> AFlexStr {
     // NOTE: We have a disadvantage to `String` because we cannot call `estimated_capacity()` on args
     // As such, we cannot assume a given needed capacity - we start with a stack allocated buffer
     // and only promote to a heap buffer if a write won't fit
-    let mut builder = build::FlexStrBuilder::new();
+    let mut builder = builder::FlexStrBuilder::new();
     builder
         .write_fmt(args)
         .expect("a formatting trait implementation returned an error");
